/*    Clock Firmware for the ESP32 series of boards and nixie clock PN Z57xM6DV2 available at www.neonixie.com/Z57xM6DV2/
 *    Copyright © 2022 Neonixie.com, All Spectrum
 *    Permission for use and modification is provided for clocks purchased from us.
 *    Software is provided as-is, with no warranty expressed or implied.
 *    
 *    YOU SHOULD NOT PUT THIS DEVICE ON THE PUBLIC INTERNET - KEEP IT BEHIND A ROUTER/FIREWALL!
 *    
 *    Auth info
 *    - Web based firmware update, http username = "z57clock"   password = "neonixie.com";
 *    - Arduino OTA firmware update, username = "neonixie.com"
 *    - mDNS hostname, usually used with Arduino OTA = Z57clock-MAC_ADDRESS (Z57clock-MAC_ADDRESS.local)
 *    - SSID when clock is in AP mode/SETUP mode (when no available access points to connect to), SSID = "z57clock"   password = "neonixie.com"
 *    
 *    Revision History
 *    0.90  07-31-2022  First public release
 *    
 *    0.91  not final   Add Cathode poisoning prevention options
 *                      Add date display options
 *                      Add leading zero blanking option
 *                      Add AutoConnect OTA Web based uploading of new firmware
 *                      Add temperature and humidity compensation options
 *                      Add temperature display every X seconds and duration options, removed temperature display tickbox
 *                      Add correct temperature display code for degree F at or above 100
 *                      Moved struct tm timeinfo; to GLOBAL
 *                      Add total uptime in days and total cumulative runtime in days (stored in non-volatile flash)
 *                      Add temperature, humidity, lux, version, uptime and total runtime to main web page on device
 *                      Code cleanup, move original time code (before timezone support) to extra-code.txt
 *                      Changed bootup IP address display to full IP address (prior code only displayed last two octets)
 *                      Added code to reset permanent (flash / like eeprom) variables in case of corruption or incorrect values cause a crash that we cannot recover from
 *                        This is done by grounding pin 15 on the ESP while the board is reset/powered up. The display will flash all 0's three times to indicate 
 *                        that flash variables have been reset to default values. TOTAL_RUNTIME is preserved.
 *                      Added watchdog timer of 300 seconds. Mainly to force the clock to attempt to reconnect to wifi when in soft AP mode. Should not get triggered in any other mode.
 *                      Added config.autoReconnect = true; to make wifi attempt a reconnect to known routers. Without this if the wifi router went offline, the clock would not reconnect.
 *                      BETA4, Removed a return from initTime, was causing timezone not to set sometimes at initial bootup
 *                      BETA4, Removed free memory display on main http page
 *                      BETA4, Changed light sensor default to off in eeprom
 *                      BETA4, Added default brightness variable in setup(), without this if light sensor was selected and none was present display would be blank
 *                      BETA4, Change default sleep time to display IP address from 2s to 4s
 *                      BETA4, Removed extra getLocalTime function and a few lines around it in cathodepp() function
 *                      BETA4, Modified ArduinoOTA.setHostname to shorten local hostname, was full MAC now Z57clock-[last-two-octets-of-MAC]
 */
// WIFI CONFIG / CONNECTION / WEB SERVER BASED on https://hieromon.github.io/AutoConnect/
// NTP BASED on https://RandomNerdTutorials.com/esp8266-nodemcu-date-time-ntp-client-server-arduino/
// EEPROM / "PREFERENCES" https://randomnerdtutorials.com/esp32-save-data-permanently-preferences/

//  THIS IS WHAT IS NORMALLY USED FOR DEV using original esp32-devkitc-ve BOARD, Also works with ESP32-DevKitC-32D
// Board Node32s    or    DOIT ESP32 DEVKIT V1  seems to work too! // upload speed 921600 // flash freq 80mhz // part scheme default

const char VERSION[] = "0.91 BETA4";
const char compile_date[] = __DATE__ " " __TIME__;

#include <WiFi.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <Wire.h>
#include <BH1750.h>
#include <Preferences.h>    // FOR "EEPROM" SAVING, IT'S CALLED "Preferences" in Arduino
Preferences preferences;

#include "z57xm6d-definitions.h"    // Mapping for HV5122 shift registers.. SEC_ONE_MAP, SEC_TEN_MAP, etc...

// OTA STUFF
#include <ArduinoOTA.h>
#include <ESPmDNS.h>

// Watchdog timer
#include <esp_task_wdt.h>

// AHT20 CODE BASED ON https://github.com/adafruit/Adafruit_AHTX0
// INSTALL FROM LIBRARY MANAGER AND SEARCH FOR AHT20, install dependencies as well
#include <Adafruit_AHTX0.h>
Adafruit_AHTX0 aht;

struct tm timeinfo;

float HUMIDITY;
float TEMPERATURE_C;
float TEMPERATURE_F;

float lux;

int SETTING_BRIGHTNESS;
int SETTING_LUXMETER;
int SETTING_12H;
int SETTING_LZBLANK;
int SETTING_DATEFORMAT;
int SETTING_DATEDISP_SEC;
int SETTING_DATEDISP_DUR;
String SETTING_TIMEZONE;
//int SETTING_DISPLAYTEMP;
int SETTING_TEMP_IN_C;
int SETTING_TEMPDISP_SEC;
int SETTING_TEMPDISP_DUR;
int SETTING_TEMP_COMP;
int SETTING_HUMIDITY_COMP;
int SETTING_CATHODEPP;
int SETTING_CATHODEPPSTART;
int SETTING_CATHODEPPDUR;
int SETTING_CATHODEPPREV;
int SETTING_CATHODEPPREVD;

int BRIGHTNESS;

float TOTAL_RUNTIME;   // TOTAL TIME CLOCK HAS BEEN POWERED UP IN DAYS
int TOTAL_RUNTIME_SECONDS_COUNTER = 0;
float UPTIME = 0;       // UPTIME IN DAYS

BH1750 lightMeter;

// FOR TIMEZONE???  // FROM     https://randomnerdtutorials.com/esp32-ntp-timezones-daylight-saving/
#include "time.h"

// C:\Users\$YOUR-USERNAME\Documents\Arduino\libraries\AutoConnect\src\AutoConnectDefs.h
// Uncomment the following AC_DEBUG to enable debug output. IN THE ABOVE FILE!!!!!!!!
// #define AC_DEBUG ////////////// IN THE ABOVE FILE!!!!!!!!
#include <AutoConnect.h>  // CHECK http://IP/_ac  when connected!
#include <WebServer.h>
WebServer Server;
AutoConnect Portal(Server);
AutoConnectConfig config("z57clock", "neonixie.com");    // This is the default SSID and PASSWORD used when the device can't connect and goes to "AP" mode

void rootPage() {
  //char content[] = "Hello, This is a nixie clock. Model Z57xM6DV2 from www.neonixie.com!<br>\
To configure settings click <a href=/clockconfig>here</a><br>Online documentation is available <a href=http://www.neonixie.com/Z57XM6DV2/?s=fromaclock>Here</a><br>";
  ////////////////////////////////////////
  // MAKE SURE contents is big enough!  //
  ////////////////////////////////////////
  char content[800];

  sprintf(content, "Hello, This is a nixie clock. Model Z57xM6DV2 from www.neonixie.com!<br>"\
  "To configure settings click <a href=/clockconfig>here</a><br>"\
  "Online documentation is available <a href=http://www.neonixie.com/Z57XM6DV2/?s=fromaclock>Here</a><br>"\
  "<br>"\
  "Temp C: %.2f Temp F: %.2f Humidity: %.0f<br>"\
  "Lux: %.0f<br>"\
  "Version %s compile time: %s<br>"\
  "Uptime Seconds: %lld  Uptime Days: %.2f  Total Runtime Days: %.2f"\
  , TEMPERATURE_C, TEMPERATURE_F, HUMIDITY, lux, VERSION, compile_date, esp_timer_get_time()/1000000, UPTIME, TOTAL_RUNTIME);

  // removed in 0.91 BETA4... "Free memory: %d" ...  ESP.getFreeHeap() 
  
  Server.send(200, "text/html", content);
}

// Define NTP Client to get time
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, "pool.ntp.org");

int sec_ones, sec_tens, min_ones, min_tens, hour_ones, hour_tens;

int previous_tm_sec;  // USE THIS TO SEE WHEN A SECOND TICKS OVER

int PREVIOUS_TEMPERATURE_READ_SECONDS;

int OE_PIN = 19;
int CLOCK_PIN = 18;
int DATA_PIN = 4;

// ADD PAGE TEST
const char PAGE_ADD[] PROGMEM = R"(
{
  "uri": "/clockconfig",
  "title": "Clock Config",
  "menu": true,
  "element": [
    {
      "name": "SETTING_BRIGHTNESS",
      "type": "ACInput",
      "label": "Brightness 0-255",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_LUXMETER",
      "type": "ACCheckbox",
      "value": "1",
      "label": "Enable LUX Meter, Overrides brightness",
      "checked": "true",
      "global": "true"
    },
    {
      "name": "SETTING_TEMP_IN_C",
      "type": "ACCheckbox",
      "value": "1",
      "label": "Temperature, set to Celsius - default is Fahrenheit",
      "checked": "false",
      "global": "true"
    },
    {
      "name": "SETTING_TEMPDISP_SEC",
      "type": "ACInput",
      "label": "Temperature display, every X seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_TEMPDISP_DUR",
      "type": "ACInput",
      "label": "Temperature Display, duration in seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_TEMP_COMP",
      "type": "ACInput",
      "label": "Temperature compensation in 1/10th degree C",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_HUMIDITY_COMP",
      "type": "ACInput",
      "label": "Humidity compensation",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_12H",
      "type": "ACCheckbox",
      "value": "1",
      "label": "Enable 12 hour display",
      "checked": "false",
      "global": "true"
    },
    {
      "name": "SETTING_LZBLANK",
      "type": "ACCheckbox",
      "value": "1",
      "label": "Enable leading zero blanking - time only",
      "checked": "false",
      "global": "true"
    },
    {
      "name": "SETTING_DATEFORMAT",
      "type": "ACSelect",
      "option":  [
          "Month Day Year - default",
          "Year Month Day",
          "Day Month Year"
        ],
      "label": "Date Format"
    },
    {
      "name": "SETTING_DATEDISP_SEC",
      "type": "ACInput",
      "label": "Date Display, every X seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_DATEDISP_DUR",
      "type": "ACInput",
      "label": "Date Display, duration in seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_CATHODEPP",
      "type": "ACSelect",
      "option":  [
          "Disable - default",
          "Enable",
          "Enable - override brightness setting to FULL",
          "Enable - override brightness setting to HALF"
        ],
      "label": "Cathode Poisoning Prevention"
    },
    {
      "name": "SETTING_CATHODEPPSTART",
      "type": "ACInput",
      "label": "Cathode Poisoning Prevention Start Hour 0-23 - default 3",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_CATHODEPPDUR",
      "type": "ACInput",
      "label": "Cathode Poisoning Prevention Duration 1-24 hours - default 1",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_CATHODEPPREV",
      "type": "ACInput",
      "label": "Cathode Poisoning Prevention, Revert to regular display; every X seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_CATHODEPPREVD",
      "type": "ACInput",
      "label": "Cathode Poisoning Prevention, Revert to regular display; duration seconds",
      "apply": "number",
      "global": "true"
    },
    {
      "name": "SETTING_TIMEZONE",
      "type": "ACInput",
      "label": "Timezone",
      "apply": "text",
      "global": "true"
    },
    {
      "name": "sometimezonetext",
      "type": "ACText",
      "value": "Example zones:<br>UTC0<br>PST8PDT,M3.2.0,M11.1.0<br>MST7<br>MST7MDT,M3.2.0,M11.1.0<br>CST6CDT,M3.2.0,M11.1.0<br>EST5EDT,M3.2.0,M11.1.0<br>More Timezone codes <a href=http:/\/github.com/nayarsystems/posix_tz_db/blob/master/zones.csv>Here</a><br><br>",
      "style": ""
    },
    {
      "name": "somemoretextbla1",
      "type": "ACText",
      "value": "",
      "style": ""
    },
    {
      "name": "add",
      "type": "ACSubmit",
      "value": "Submit",
      "uri": "/clockconfig"
    }
  ]
}
)";

//AutoConnect portal;
AutoConnectAux  page_add;
//AutoConnectAux  page_results;

int INITIALIZE_WEB_VARIABLES=0;

String onclockconfig(AutoConnectAux& aux, PageArgument& args) {
  if ( INITIALIZE_WEB_VARIABLES == 0 ) {
    aux["SETTING_BRIGHTNESS"].as<AutoConnectInput>().value = SETTING_BRIGHTNESS;
    aux["SETTING_LUXMETER"].as<AutoConnectCheckbox>().checked = SETTING_LUXMETER;
    
    //aux["SETTING_DISPLAYTEMP"].as<AutoConnectCheckbox>().checked = SETTING_DISPLAYTEMP;
    aux["SETTING_TEMP_IN_C"].as<AutoConnectCheckbox>().checked = SETTING_TEMP_IN_C;
    aux["SETTING_TEMPDISP_SEC"].as<AutoConnectInput>().value = SETTING_TEMPDISP_SEC;
    aux["SETTING_TEMPDISP_DUR"].as<AutoConnectInput>().value = SETTING_TEMPDISP_DUR; 
    aux["SETTING_TEMP_COMP"].as<AutoConnectInput>().value = SETTING_TEMP_COMP;
    aux["SETTING_HUMIDITY_COMP"].as<AutoConnectInput>().value = SETTING_HUMIDITY_COMP;
    
    aux["SETTING_12H"].as<AutoConnectCheckbox>().checked = SETTING_12H;
    aux["SETTING_LZBLANK"].as<AutoConnectCheckbox>().checked = SETTING_LZBLANK;

    aux["SETTING_DATEFORMAT"].as<AutoConnectSelect>().selected = SETTING_DATEFORMAT;
    aux["SETTING_DATEDISP_SEC"].as<AutoConnectInput>().value = SETTING_DATEDISP_SEC;
    aux["SETTING_DATEDISP_DUR"].as<AutoConnectInput>().value = SETTING_DATEDISP_DUR;
    
    aux["SETTING_CATHODEPP"].as<AutoConnectSelect>().selected = SETTING_CATHODEPP;
    aux["SETTING_CATHODEPPSTART"].as<AutoConnectInput>().value = SETTING_CATHODEPPSTART;
    aux["SETTING_CATHODEPPDUR"].as<AutoConnectInput>().value = SETTING_CATHODEPPDUR;
    aux["SETTING_CATHODEPPREV"].as<AutoConnectInput>().value = SETTING_CATHODEPPREV;
    aux["SETTING_CATHODEPPREVD"].as<AutoConnectInput>().value = SETTING_CATHODEPPREVD;
    
    aux["SETTING_TIMEZONE"].as<AutoConnectInput>().value = SETTING_TIMEZONE;
    aux["somemoretextbla1"].as<AutoConnectText>().value = "Version " + String(VERSION) + " compile time: " + String(compile_date) + "<br>";   
    INITIALIZE_WEB_VARIABLES = 1;
  } else {

    preferences.begin("clock-settings", false);
    
    SETTING_BRIGHTNESS = aux["SETTING_BRIGHTNESS"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("BRIGHTNESS", SETTING_BRIGHTNESS);
    
    SETTING_LUXMETER = aux["SETTING_LUXMETER"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("LUXMETER", SETTING_LUXMETER);

    //SETTING_DISPLAYTEMP = aux["SETTING_DISPLAYTEMP"].as<AutoConnectCheckbox>().checked;
    //preferences.putInt("DISPLAYTEMP", SETTING_DISPLAYTEMP);
    SETTING_TEMP_IN_C = aux["SETTING_TEMP_IN_C"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("TEMP_IN_C", SETTING_TEMP_IN_C); 
    SETTING_TEMPDISP_SEC = aux["SETTING_TEMPDISP_SEC"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMPDISP_SEC", SETTING_TEMPDISP_SEC);
    SETTING_TEMPDISP_DUR = aux["SETTING_TEMPDISP_DUR"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMPDISP_DUR", SETTING_TEMPDISP_DUR);  
    
    SETTING_TEMP_COMP = aux["SETTING_TEMP_COMP"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("TEMP_COMP", SETTING_TEMP_COMP);
    SETTING_HUMIDITY_COMP = aux["SETTING_HUMIDITY_COMP"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("HUMIDITY_COMP", SETTING_HUMIDITY_COMP);

    SETTING_12H = aux["SETTING_12H"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("12H", SETTING_12H);

    SETTING_LZBLANK = aux["SETTING_LZBLANK"].as<AutoConnectCheckbox>().checked;
    preferences.putInt("LZBLANK", SETTING_LZBLANK);

    SETTING_DATEFORMAT = aux["SETTING_DATEFORMAT"].as<AutoConnectSelect>().selected;
    preferences.putInt("DATEFORMAT", SETTING_DATEFORMAT);
    SETTING_DATEDISP_SEC = aux["SETTING_DATEDISP_SEC"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("DATEDISP_SEC", SETTING_DATEDISP_SEC);
    SETTING_DATEDISP_DUR = aux["SETTING_DATEDISP_DUR"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("DATEDISP_DUR", SETTING_DATEDISP_DUR);    

    SETTING_CATHODEPP = aux["SETTING_CATHODEPP"].as<AutoConnectSelect>().selected;
    preferences.putInt("CATHODEPP", SETTING_CATHODEPP);
    SETTING_CATHODEPPSTART = aux["SETTING_CATHODEPPSTART"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("CATHODEPPSTART", SETTING_CATHODEPPSTART);
    SETTING_CATHODEPPDUR = aux["SETTING_CATHODEPPDUR"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("CATHODEPPDUR", SETTING_CATHODEPPDUR);
    SETTING_CATHODEPPREV = aux["SETTING_CATHODEPPREV"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("CATHODEPPREV", SETTING_CATHODEPPREV);
    SETTING_CATHODEPPREVD = aux["SETTING_CATHODEPPREVD"].as<AutoConnectInput>().value.toInt();
    preferences.putInt("CATHODEPPREVD", SETTING_CATHODEPPREVD);

    SETTING_TIMEZONE = aux["SETTING_TIMEZONE"].as<AutoConnectInput>().value;
    preferences.putString("TIMEZONE", SETTING_TIMEZONE);
    
    setTimezone(SETTING_TIMEZONE);
    
    preferences.end();
  }

  return String();
}

String onResults(AutoConnectAux& aux, PageArgument& args) {
  int valueA = args.arg("valueA").toInt();
  int valueB = args.arg("valueB").toInt();

  ledcWrite(0, valueA);

  aux["results"].as<AutoConnectText>().value = String(valueA) + " + " + String(valueB) + " = " + String(valueA + valueB);
  return String();
}
// END ADD PAGE TEST


void WRITE_BITS(int x) {
  int var = 9;
  while (var >= 0) {
    digitalWrite(CLOCK_PIN, 0);
    //digitalWrite(DATA_PIN, (x >> var) & 0x01);    // REGULAR
    digitalWrite(DATA_PIN, !((x >> var) & 0x01));      // INVERTED
    digitalWrite(CLOCK_PIN, 1);
    var--;
  }
}

void WRITE_ONELOWBIT() {
    digitalWrite(CLOCK_PIN, 0);
    digitalWrite(DATA_PIN, 1);
    digitalWrite(CLOCK_PIN, 1);
}

void WRITE_ONEHIGHBIT() {
    digitalWrite(CLOCK_PIN, 0);
    digitalWrite(DATA_PIN, 0);
    digitalWrite(CLOCK_PIN, 1);
}

void WRITE_DISPLAY(int HT, int HO, int MT, int MO, int ST, int SO)
{
    WRITE_BITS(MIN_TEN_MAP[MT]);

    if ( SO % 2 == 0 ) {
      WRITE_ONELOWBIT();    // NEON BULB
    } else {
      WRITE_ONEHIGHBIT();    // NEON BULB
    }
    
    WRITE_BITS(HOUR_ONE_MAP[HO]);
    WRITE_BITS(HOUR_TEN_MAP[HT]);
    
    WRITE_ONELOWBIT();    // PIN 35 U4 /  NO CONNECTION

    WRITE_BITS(SEC_ONE_MAP[SO]);
    WRITE_BITS(SEC_TEN_MAP[ST]);
    
    if ( SO % 2 == 0 ) {
      WRITE_ONELOWBIT();    // NEON BULB
    } else {
      WRITE_ONEHIGHBIT();    // NEON BULB
    } 
    
    WRITE_BITS(MIN_ONE_MAP[MO]);
    
    WRITE_ONELOWBIT();    // PIN 35 NO CONNECTION  

    digitalWrite(CLOCK_PIN, 0);   // WRITE CLOCK PIN TO LOW TO KEEP LOGIC LEVEL SHIFTER (RESISTOR AND FET) TEMPERATURES LOWER
    digitalWrite(DATA_PIN, 0);    // WRITE DATA PIN TO LOW TO KEEP LOGIC LEVEL SHIFTER (RESISTOR AND FET) TEMPERATURES LOWER
}

void demo_digits(){
    int var = 0;
    while (var <= 9) {
      WRITE_DISPLAY(var, var, var, var, var, var);
      var++;
      delay(500);      
    }
}

void setTimezone(String timezone){
  Serial.printf(" Setting Timezone to %s\n",timezone.c_str());
  setenv("TZ",timezone.c_str(),1);  //  Now adjust the TZ.  Clock settings are adjusted to show the new local time
  tzset();
}

void initTime(String timezone){
  //struct tm timeinfo; // Moved to global

  Serial.println("Setting up time");
  configTime(0, 0, "pool.ntp.org");    // First connect to NTP server, with 0 TZ offset
  if(!getLocalTime(&timeinfo)){
    Serial.println(" Failed to obtain time");
    //return; // REMOVED AS OF 0.91 BETA4, this was at times causing timezone to not set on initial bootup
  } else {
    Serial.println(" Got the time from NTP");
  }
  setTimezone(timezone);
}

void printLocalTime(){
  //struct tm timeinfo; // Moved to global
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time 1");
    return;
  }
  Serial.println(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");
}

void setTime(int yr, int month, int mday, int hr, int minute, int sec, int isDst){
  struct tm tm;

  tm.tm_year = yr - 1900;   // Set date
  tm.tm_mon = month-1;
  tm.tm_mday = mday;
  tm.tm_hour = hr;      // Set time
  tm.tm_min = minute;
  tm.tm_sec = sec;
  tm.tm_isdst = isDst;  // 1 or 0
  time_t t = mktime(&tm);
  Serial.printf("Setting time: %s", asctime(&tm));
  struct timeval now = { .tv_sec = t };
  settimeofday(&now, NULL);
}

void cathodepp (void)
{
  if ( (timeinfo.tm_hour >= SETTING_CATHODEPPSTART && timeinfo.tm_hour < (SETTING_CATHODEPPSTART+SETTING_CATHODEPPDUR)) || timeinfo.tm_hour < ((SETTING_CATHODEPPSTART+SETTING_CATHODEPPDUR)-24) ) {
    if ( SETTING_CATHODEPP == 3 ) {
       //ledcWrite(0, 0);     // SET TO FULL BRIGHTNESS
       BRIGHTNESS=255;
    }
    if ( SETTING_CATHODEPP == 4 ) {
       //ledcWrite(0, 128);   // SET TO HALF BRIGHTNESS
       BRIGHTNESS=128;
    }

    //if ( timeinfo.tm_sec == 0 || timeinfo.tm_sec == 10 || timeinfo.tm_sec == 20 || timeinfo.tm_sec == 30 || timeinfo.tm_sec == 40 || timeinfo.tm_sec == 50 ) { // OLD REMOVED 0.91 BETA4
    if ( timeinfo.tm_sec%SETTING_CATHODEPPREV < SETTING_CATHODEPPREVD ) {   // NEW 0.91 BETA4 Check if we need to revert to a regular display every SETTING_CATHODEPPREV (SECONDS) for SETTING_CATHODEPPREVD (DURATION)
      //juggleoutput(0);
    } else {
      sec_ones=timeinfo.tm_min/6;
      sec_tens=timeinfo.tm_min/6;
      min_ones=timeinfo.tm_min/6;
      min_tens=timeinfo.tm_min/6;
      hour_ones=timeinfo.tm_min/6;
      hour_tens=timeinfo.tm_min/6;
    }
              
  }
}

void display_temperature() {
  
  if ( SETTING_TEMPDISP_SEC != 0 ) {
    if ( timeinfo.tm_sec%10 == 0 ) {    // READ TEMP AND HUMIDITY EVERY 10 SECONDS
      //aht.begin();  // THIS IS MOVED OUT OF THE REGULAR LOOP
      sensors_event_t aht_humidity, aht_temp;
      aht.getEvent(&aht_humidity, &aht_temp);// populate temp and humidity objects with fresh data
      HUMIDITY=aht_humidity.relative_humidity + SETTING_HUMIDITY_COMP;
      TEMPERATURE_C=aht_temp.temperature + (SETTING_TEMP_COMP/float(10));
      TEMPERATURE_F=(TEMPERATURE_C * 9/5) + 32;
    }
    //if ( (timeinfo.tm_sec%SETTING_TEMPDISP_SEC <= SETTING_TEMPDISP_DUR) || ((timeinfo.tm_sec%SETTING_TEMPDISP_SEC)-1) == 0 ) {
    if ( timeinfo.tm_sec%SETTING_TEMPDISP_SEC < SETTING_TEMPDISP_DUR ) {
      if ( SETTING_TEMP_IN_C == 1 ) {
          hour_tens=int(TEMPERATURE_C) / 10;
          hour_ones=int(TEMPERATURE_C) % 10;;
          min_tens=int(TEMPERATURE_C * 10) %10;
          min_ones=10;      
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
      } else {
        if ( TEMPERATURE_F >= 100 ) {          // 100+F display as xxx and round
          hour_tens=int(round(TEMPERATURE_F)) %1000/100;
          hour_ones=int(round(TEMPERATURE_F)) %100/10;
          min_tens=int(round(TEMPERATURE_F)) % 10;
          min_ones=10;
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
        } else {                              // LESS THEN 100F display as xx.x  
          hour_tens=int(TEMPERATURE_F) / 10;
          hour_ones=int(TEMPERATURE_F) % 10;;
          min_tens=int(TEMPERATURE_F * 10) %10;
          min_ones=10;      
          sec_tens=int(HUMIDITY) / 10;
          sec_ones=int(HUMIDITY) % 10;
        }
      }
    }
    Serial.print("Temp: "); Serial.print(TEMPERATURE_C); Serial.print(" C "); Serial.print(TEMPERATURE_F); Serial.print(" F "); Serial.print("Humidity: "); Serial.print(HUMIDITY); Serial.print("% rH ");
  }
}

void display_date() {
  if ( SETTING_DATEDISP_SEC != 0 ) {
    if ( timeinfo.tm_sec%SETTING_DATEDISP_SEC < SETTING_DATEDISP_DUR ) {
      switch (SETTING_DATEFORMAT) {
         case 2:
          sec_ones = timeinfo.tm_mday % 10;
          sec_tens = timeinfo.tm_mday / 10;
          min_ones = (timeinfo.tm_mon+1) % 10;
          min_tens = (timeinfo.tm_mon+1) / 10;
          hour_ones = (timeinfo.tm_year-100) % 10;
          hour_tens = (timeinfo.tm_year-100) / 10;
          break;
         case 3:
          sec_ones = (timeinfo.tm_year-100) % 10;
          sec_tens = (timeinfo.tm_year-100) / 10;
          min_ones = (timeinfo.tm_mon+1) % 10;
          min_tens = (timeinfo.tm_mon+1) / 10;
          hour_ones = timeinfo.tm_mday % 10;
          hour_tens = timeinfo.tm_mday / 10;
          break;          
         default: // THIS IS CASE 1, our default case
          sec_ones = (timeinfo.tm_year-100) % 10;
          sec_tens = (timeinfo.tm_year-100) / 10;
          min_ones = timeinfo.tm_mday % 10;
          min_tens = timeinfo.tm_mday / 10;
          hour_ones = (timeinfo.tm_mon+1) % 10;
          hour_tens = (timeinfo.tm_mon+1) / 10;
          break;
      }
    }
  }

}

void setup() {
  esp_task_wdt_init(300, true);  // ENABLE WATCHDOG
  esp_task_wdt_add(NULL);       // ADD CURRENT THREAD
  
  pinMode(CLOCK_PIN, OUTPUT);
  pinMode(DATA_PIN, OUTPUT);

  // SET UP PWM ON OUTPUT ENABLE PIN FOR DIMMING
  ledcSetup(0, 400, 8);   //  ledcSetup(uint8_t channel, double freq, uint8_t resolution_bits);
  ledcAttachPin(OE_PIN, 0);
  ledcWrite(0, 0);    // INVERTED (0, 0)   // REGULAR (0, 255)

  BRIGHTNESS=255;   // ADDED in 0.91v BETA4, without this when the light meter was enabled, but no light meter was installed, the brightness was never set and display would be blank

  // CLEAR OUTPUT BITS OF SHIFT REGISTER
  int var = 0;
  while (var <= 65) {
    WRITE_ONELOWBIT();    // NO NEED TO WORRY ABOUT INVERTING HERE, IT'S DONE IN THE FUNCTION BOTH LOWBIT AND HIGHBIT
    var++;
  }

  // SETTINGS THAT GET STORED IN FLASH
  preferences.begin("clock-settings", false);

  TOTAL_RUNTIME = preferences.getFloat("TOTAL_RUNTIME", 0);   // Get TOTAL_RUNTIME before processing possible reset of variables below

  pinMode(15, INPUT_PULLUP);           // READ PIN 15 and if low (ground) reset permanent (eeprom/flash) variables to their default values, except TOTAL_RUNTIME
  if ( digitalRead(15) == 0 ) {
    preferences.clear();
    WRITE_DISPLAY(0,0,0,0,0,0); delay(500); WRITE_DISPLAY(10,10,10,10,10,10); delay(500);
    WRITE_DISPLAY(0,0,0,0,0,0); delay(500); WRITE_DISPLAY(10,10,10,10,10,10); delay(500);
    WRITE_DISPLAY(0,0,0,0,0,0); delay(500); WRITE_DISPLAY(10,10,10,10,10,10); delay(500);
  }

  preferences.putFloat("TOTAL_RUNTIME", TOTAL_RUNTIME);   // Put TOTAL_RUNTIME back into permanent storage, as we don't want to reset this value

  SETTING_BRIGHTNESS = preferences.getInt("BRIGHTNESS", 255);
  SETTING_LUXMETER = preferences.getInt("LUXMETER", 0);
  //SETTING_DISPLAYTEMP = preferences.getInt("DISPLAYTEMP", 0);
  SETTING_TEMPDISP_SEC = preferences.getInt("TEMPDISP_SEC", 0);
  SETTING_TEMPDISP_DUR = preferences.getInt("TEMPDISP_DUR", 2);
  
  SETTING_TEMP_IN_C = preferences.getInt("TEMP_IN_C", 0);
  SETTING_TEMP_COMP = preferences.getInt("TEMP_COMP", 0);
  SETTING_HUMIDITY_COMP = preferences.getInt("HUMIDITY_COMP", 0);
  SETTING_12H = preferences.getInt("12H", 0);
  SETTING_LZBLANK = preferences.getInt("LZBLANK", 0);

  SETTING_DATEFORMAT = preferences.getInt("DATEFORMAT", 1);
  SETTING_DATEDISP_SEC = preferences.getInt("DATEDISP_SEC", 0);
  SETTING_DATEDISP_DUR = preferences.getInt("DATEDISP_DUR", 2);
    
  SETTING_CATHODEPP = preferences.getInt("CATHODEPP", 1);
  SETTING_CATHODEPPSTART = preferences.getInt("CATHODEPPSTART", 3);
  SETTING_CATHODEPPDUR = preferences.getInt("CATHODEPPDUR", 1);
  SETTING_CATHODEPPREV = preferences.getInt("CATHODEPPREV", 10);
  SETTING_CATHODEPPREVD = preferences.getInt("CATHODEPPREVD", 4);
  
  SETTING_TIMEZONE = preferences.getString("TIMEZONE", "PST8PDT,M3.2.0,M11.1.0");
  
  preferences.end();

  // ORIGINAL SEC_ONE_MAP[], SEC_TEN_MAP[], and so on... HV5x222 to Nixie Tube digit DEFINITIONS WHERE HERE, NOW IN z57xm6d-definitions.h 

  demo_digits();
  
  // Initialize Serial Monitor
  Serial.begin(115200);
  Serial.println();
  Server.on("/", rootPage);
  config.hostName = "z57clock";  

  config.auth = AC_AUTH_DIGEST;
  config.authScope = AC_AUTHSCOPE_PARTIAL;    // This configures authentication only on the web based firmware update html page (autoconnect.h)
  config.username = "z57clock";
  config.password = "neonixie.com";

  config.ota = AC_OTA_BUILTIN;  // Turns on web based firmware update (built into autoconnect.h)

  config.autoReconnect = true;    // Without this option, when the wifi router was shut off and turned back on, the clock would not attempt a connection again to the wifi, it would just go into soft AP mode. Adding this fixed it and it now attempts to reconnect when the wifi router is back online. This was clued in from https://github.com/Hieromon/AutoConnect/issues/80
  Portal.config(config);
  if (Portal.begin()) {
    Serial.println("HTTP server:" + WiFi.localIP().toString());    
  }

  // Initialize a NTPClient to get time   // THIS IS THE PRIOR OLDER STUFF, NO LONGER USED REALLY
  timeClient.begin();
  timeClient.setTimeOffset(-28800);   // Set offset time in seconds to adjust for your timezone, for example: GMT +1 = 3600     GMT +8 = 28800    GMT -1 = -3600    GMT 0 = 0

// ADD PAGE TEST
  delay(1000);
  page_add.load(PAGE_ADD);
  //page_results.load(PAGE_RESULTS);
  //Portal.join({ page_add, page_results });
  Portal.join({ page_add });
  Portal.on("/clockconfig", onclockconfig);
  //Portal.on("/results", onResults);
  Portal.begin();
// END ADD PAGE TEST


  // DISPLAY LOCAL IP ADDRESS
  WRITE_DISPLAY(WiFi.localIP()[0]/100, WiFi.localIP()[0]%100/10, WiFi.localIP()[0]%10, WiFi.localIP()[1]/100, WiFi.localIP()[1]%100/10, WiFi.localIP()[1]%10);
  delay(4000);
  WRITE_DISPLAY(WiFi.localIP()[2]/100, WiFi.localIP()[2]%100/10, WiFi.localIP()[2]%10, WiFi.localIP()[3]/100, WiFi.localIP()[3]%100/10, WiFi.localIP()[3]%10);
  delay(4000);    

  initTime(SETTING_TIMEZONE);   // PST8PDT,M3.2.0,M11.1.0   // THIS IS THE ACTUAL ONE THAT WORKS WITH TIMEZONES NOW 06/04/2022
  printLocalTime();

  // OTA STUFF START

  // Port defaults to 3232
  // ArduinoOTA.setPort(3232);

  // Hostname defaults to esp3232-[MAC]
  // ArduinoOTA.setHostname("myesp32");

  char tmp_hostname[30];
  uint8_t mac[6];
  WiFi.macAddress(mac);
  sprintf(tmp_hostname, "Z57clock-%02x%02x", mac[4], mac[5]);   // BETA4 REMOVED mac[0] - mac[4] to shorten hostname
  ArduinoOTA.setHostname(tmp_hostname);

  // No authentication by default
  // ArduinoOTA.setPassword("admin");
  ArduinoOTA.setPassword("neonixie.com");

  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");

        ArduinoOTA
    .onStart([]() {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)
        type = "sketch";
      else // U_SPIFFS
        type = "filesystem";

      // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
      Serial.println("Start updating " + type);
    })
    .onEnd([]() {
      Serial.println("\nEnd");
    })
    .onProgress([](unsigned int progress, unsigned int total) {
      Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    })
    .onError([](ota_error_t error) {
      Serial.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");
    });

  ArduinoOTA.begin();
  // OTA STUFF END

  Wire.begin();
  lightMeter.begin();

  // START AHT20 TEMP HUMIDITY SENSOR
  aht.begin();
}

// WIFI RECONNECT TEST // ADDED 01/10/2022, to see if it fixes reconnect problem. // 07/11/2022 SEEMS TO RECONNECT WHEN NEEDED
unsigned long previousMillis = 0;
unsigned long interval = 30000;

unsigned long NTP_previousMillis = 9999999;   // MUST BE LARGER THEN interval
unsigned long NTP_interval = 300000;   // INTERVAL IN MILLISECONDS TO RUN THE timeClient.update() FUNCTION.  60000 = 1 minute  300000 = 5 minutes

void loop() {
  
  // NEW STUFF WITH TIMEZONE
  //printLocalTime();
  //struct tm timeinfo;   // MOVED TO GLOBAL
  if(!getLocalTime(&timeinfo)){
    Serial.println("Failed to obtain time 1");
    return;
  }
  
  ArduinoOTA.handle();    // OTA
  
  Portal.handleClient();
  
  unsigned long NTP_currentMillis = millis();
  if ( NTP_currentMillis - NTP_previousMillis >=NTP_interval ) {
    Serial.print("timeClient.update start");
    timeClient.update();      // NOTE, POSSIBLY DUE TO DNS DELAYS OR ?? THIS FUNCTION WILL SOMETIMES DELAY BY SEVERAL SECONDS. SET TO RUN ONLY PERIODICALLY SO AS NOT TO DISTRUPT THE CLOCK DISPLAY
    Serial.print("timeClient.update end");
    NTP_previousMillis = NTP_currentMillis;
  }

  sec_ones = timeinfo.tm_sec % 10;
  sec_tens = timeinfo.tm_sec / 10;
  min_ones = timeinfo.tm_min % 10;
  min_tens = timeinfo.tm_min / 10;
  if ( SETTING_12H == 1 ) {
    if ( timeinfo.tm_hour == 0 ) {  // MIDNIGHT, AM
      hour_ones=2;
      hour_tens=1;
    } else if ( timeinfo.tm_hour >= 1 && timeinfo.tm_hour <= 11 ) { // 1 to 11 AM
      hour_ones = timeinfo.tm_hour % 10;
      hour_tens = timeinfo.tm_hour / 10;
    } else if ( timeinfo.tm_hour == 12 ) {  // NOON, PM
      hour_ones=2;
      hour_tens=1;
    } else {    // THIS IS ANYTHING AFTER 1300
      hour_ones = ((timeinfo.tm_hour)-12) % 10;
      hour_tens = ((timeinfo.tm_hour)-12) / 10; 
    }
  } else {
    hour_ones = timeinfo.tm_hour % 10;
    hour_tens = timeinfo.tm_hour / 10;
  }
  
  if ( SETTING_LZBLANK == 1 ) {
    if ( hour_tens == 0 ) {
      hour_tens=10;
    }
  }

// WIFI RECONNECT TEST
// ADDED 01/10/2022, to see if it fixes reconnect problem. // 05/29/2022 SEEMS TO WORK WELL
unsigned long currentMillis = millis();
// if WiFi is down, try reconnecting
if ((WiFi.status() != WL_CONNECTED) && (currentMillis - previousMillis >=interval)) {
  Serial.print(millis());
  Serial.println("Reconnecting to WiFi...");
  WiFi.disconnect();
  WiFi.reconnect();
  previousMillis = currentMillis;
}

// START ONCE PER SECOND LOOP
if ( previous_tm_sec != timeinfo.tm_sec ) { 
  previous_tm_sec=timeinfo.tm_sec;

  esp_task_wdt_reset();   // Watchdog timer reset

  TOTAL_RUNTIME_SECONDS_COUNTER++;
  if ( TOTAL_RUNTIME_SECONDS_COUNTER >= 864 ) {  // 1/100th of a day has passed, increment UPTIME and TOTAL_RUNTIME, save to permanent memory, DON'T UPDATE MEMORY TOO OFTEN due to flash wear!
    TOTAL_RUNTIME_SECONDS_COUNTER=0;
    UPTIME=UPTIME + 0.01;
    TOTAL_RUNTIME=TOTAL_RUNTIME + 0.01;
    preferences.begin("clock-settings", false);
    preferences.putFloat("TOTAL_RUNTIME", TOTAL_RUNTIME);
    preferences.end();
  }

  if ( SETTING_LUXMETER == 1 ) {  
    // READ BL1750 LIGHT METER
    lux = lightMeter.readLightLevel();
    Serial.print("Light: "); Serial.print(lux); Serial.print(" lx"); Serial.print(" ");

    // SET PWM ON OUTPUT PIN FOR DIMMING
    if ( lux < 0 ) {
      // do nothing! something is wrong! Sometimes we get a -1 error reading from the sensor
    } else if ( lux > 50 ) {
      //ledcWrite(0, 0);      // FULL BRIGHTNESS  // (0, 0) for inverted output, (0, 255) for regular
      BRIGHTNESS=255;
    } else if ( lux > 5 ) {
      //ledcWrite(0, 128);    // HALF BRIGHT      // (0, 150 for inverted output, (0, 100) for regular
      BRIGHTNESS=128;
    } else {
      //ledcWrite(0, 205);    // DIM              // (0, 200) for inverted output, (0, 50) for regular
      BRIGHTNESS=50;
    }
    //sec_ones = int(lux)%10; sec_tens = int(lux)%100/10; min_ones = int(lux)%1000/100; min_tens = int(lux)%10000/1000; hour_ones = int(lux)/10000; hour_tens = 10;   // DISPLAY LUX LEVELS
  } else {
    //ledcWrite(0, abs(SETTING_BRIGHTNESS-255));  // INVERTED BRIGHTNESS
    //BRIGHTNESS=abs(SETTING_BRIGHTNESS-255);
    BRIGHTNESS=SETTING_BRIGHTNESS;
  }

  display_temperature();
  display_date();

  Serial.print(&timeinfo, "%A, %B %d %Y %H:%M:%S zone %Z %z ");
  Serial.println("");   // PUT A FINAL RETURN CHARACTER OUT THE SERIAL PORT
 
  if ( SETTING_CATHODEPP != 1 ) {
    cathodepp();
  }

  ledcWrite(0, abs(BRIGHTNESS-255));
  WRITE_DISPLAY(hour_tens, hour_ones, min_tens, min_ones, sec_tens, sec_ones);
} 
// END ONCE PER SECOND LOOP

  //Serial.println(WiFi.RSSI());    // RSSI MONITOR
  //WRITE_DISPLAY(10, 10, 10, 0, (abs(WiFi.RSSI()))/10, int((abs(WiFi.RSSI())))%10);
  
  delay(20);
}
